import inspect
from typing import cast, Dict, Hashable, Optional

from .._internal import API
from .._internal.utils import debug_repr, FinalImmutable
from ..core import Container, DependencyInstance, Provider
from ..core.utils import DependencyDebug


@API.private
class Build(FinalImmutable):
    __slots__ = ('dependency', 'kwargs', '_hash')
    dependency: Hashable
    kwargs: dict
    _hash: int

    def __init__(self, dependency: Hashable, kwargs: dict):
        assert isinstance(kwargs, dict) and len(kwargs) > 0

        try:
            # Try most precise hash first
            _hash = hash((dependency, tuple(kwargs.items())))
        except TypeError:
            # If type error, return the best error-free hash possible
            _hash = hash((dependency, tuple(kwargs.keys())))

        super().__init__(dependency, kwargs, _hash)

    def __hash__(self):
        return self._hash

    def __repr__(self):
        return f"Build(dependency={self.dependency}, kwargs={self.kwargs})"

    def __antidote_debug_repr__(self):
        return f"{debug_repr(self.dependency)}(**{self.kwargs})"

    def __eq__(self, other):
        return (isinstance(other, Build)
                and self._hash == other._hash
                and (self.dependency is other.dependency
                     or self.dependency == other.dependency)
                and self.kwargs == other.kwargs)  # noqa


@API.private
class ServiceProvider(Provider):
    def __init__(self):
        super().__init__()
        self.__services: Dict[Hashable, bool] = dict()

    def __repr__(self):
        return f"{type(self).__name__}(services={list(self.__services.items())!r})"

    def exists(self, dependency: Hashable) -> bool:
        if isinstance(dependency, Build):
            return dependency.dependency in self.__services
        return dependency in self.__services

    def clone(self, keep_singletons_cache: bool) -> 'ServiceProvider':
        p = ServiceProvider()
        p.__services = self.__services.copy()
        return p

    def maybe_debug(self, build: Hashable) -> Optional[DependencyDebug]:
        klass = build.dependency if isinstance(build, Build) else build
        try:
            singleton = self.__services[klass]
        except KeyError:
            return None
        return DependencyDebug(debug_repr(build),
                               singleton=singleton,
                               wired=[klass])

    def maybe_provide(self, build: Hashable, container: Container
                      ) -> Optional[DependencyInstance]:
        dependency = build.dependency if isinstance(build, Build) else build
        try:
            singleton = self.__services[dependency]
        except KeyError:
            return None

        klass = cast(type, dependency)
        if isinstance(build, Build) and build.kwargs:
            instance = klass(**build.kwargs)
        else:
            instance = klass()

        return DependencyInstance(instance, singleton=singleton)

    def register(self, klass: type, *, singleton: bool = True):
        if not (isinstance(klass, type) and inspect.isclass(klass)):
            raise TypeError(f"service must be a class, not {klass!r}")
        if not isinstance(singleton, bool):
            raise TypeError(f"singleton must be a boolean, not {singleton!r}")
        self._assert_not_duplicate(klass)
        self.__services[klass] = singleton



from contextlib import contextmanager

import pytest

from antidote import service, Service, Tag, Wiring, world
from antidote._providers import ServiceProvider
from antidote.exceptions import DuplicateDependencyError


@contextmanager
def does_not_raise():
    yield


@pytest.fixture(autouse=True)
def test_world():
    with world.test.empty():
        world.provider(ServiceProvider)
        yield


def test_register_simple():
    @service
    class A:
        pass

    assert isinstance(world.get(A), A)
    # singleton by default
    assert world.get(A) is world.get(A)


def test_simple():
    class A(Service):
        pass

    assert isinstance(world.get(A), A)
    # singleton by default
    assert world.get(A) is world.get(A)


def test_register_singleton():
    @service(singleton=True)
    class Singleton:
        pass

    assert world.get(Singleton) is world.get(Singleton)

    @service(singleton=False)
    class NoScope:
        pass

    assert world.get(NoScope) != world.get(NoScope)


def test_singleton():
    class Singleton(Service):
        __antidote__ = Service.Conf(singleton=True)

    assert world.get(Singleton) is world.get(Singleton)

    class NoScope(Service):
        __antidote__ = Service.Conf(singleton=False)

    assert world.get(NoScope) != world.get(NoScope)


def test_duplicate_registration():
    with pytest.raises(DuplicateDependencyError):
        @service
        class Dummy(Service):
            pass


@pytest.mark.parametrize('cls', ['test', object(), lambda: None])
def test_invalid_class(cls):
    with pytest.raises(TypeError):
        service(cls)


@pytest.mark.parametrize(
    'kwargs,expectation',
    [
        (dict(tags=object()), pytest.raises(TypeError, match=".*tags.*")),
        (dict(tags=['test']), pytest.raises(TypeError, match=".*tags.*")),
        (dict(singleton=object()), pytest.raises(TypeError, match=".*singleton.*")),
    ]
)
def test_invalid_params(kwargs, expectation):
    with expectation:
        @service(**kwargs)
        class Dummy:
            method = None


def test_not_tags():
    with world.test.empty():
        world.provider(ServiceProvider)

        @service
        class A:
            pass

        assert isinstance(world.get(A), A)
        assert world.get(A) is world.get(A)

        class B(Service):
            pass

        assert isinstance(world.get(B), B)
        assert world.get(B) is world.get(B)

    with world.test.empty():
        world.provider(ServiceProvider)
        tag = Tag()
        with pytest.raises(RuntimeError):
            @service(tags=[tag])
            class A:
                pass

        with pytest.raises(RuntimeError):
            class B(Service):
                __antidote__ = Service.Conf(tags=[tag])


def test_invalid_conf():
    with pytest.raises(TypeError, match=".*__antidote__.*"):
        class Dummy(Service):
            __antidote__ = 1


def test_no_subclass_of_service():
    class A(Service):
        pass

    with pytest.raises(TypeError, match=".*abstract.*"):
        class B(A):
            pass


@pytest.mark.parametrize('kwargs, expectation', [
    (dict(singleton=object()), pytest.raises(TypeError, match=".*singleton.*")),
    (dict(tags=object()), pytest.raises(TypeError, match=".*tags.*")),
    (dict(tags=['dummy']), pytest.raises(TypeError, match=".*tags.*")),
    (dict(wiring=object()), pytest.raises(TypeError, match=".*wiring.*")),
])
def test_conf_error(kwargs, expectation):
    with expectation:
        Service.Conf(**kwargs)


@pytest.mark.parametrize('kwargs', [
    dict(singleton=False),
    dict(tags=(Tag(),)),
    dict(wiring=Wiring(methods=['method'])),
])
def test_conf_copy(kwargs):
    conf = Service.Conf(singleton=True, tags=[]).copy(**kwargs)
    for k, v in kwargs.items():
        assert getattr(conf, k) == v



import pytest

from antidote._internal.utils import raw_getattr


def test_raw_getattr():
    class A:
        @staticmethod
        def static():
            return 'static'

        @classmethod
        def klass(cls):
            return 'klass'

        def method(self):
            return 'method'

    class B(A):
        pass

    for (cls, with_super) in [(A, False), (B, True)]:
        method = raw_getattr(cls, 'static', with_super)
        assert method is not None
        assert isinstance(method, staticmethod)

        method = raw_getattr(cls, 'klass', with_super)
        assert method is not None
        assert isinstance(method, classmethod)

        method = raw_getattr(cls, 'method', with_super)
        assert method is not None
        assert method(1) == 'method'

        with pytest.raises(AttributeError, match=".*unknown.*"):
            raw_getattr(cls, 'unknown', with_super)



import weakref
from typing import Any, Callable, Dict, Hashable, Optional, Union

from .service import Build
from .._internal import API
from .._internal.utils import debug_repr, FinalImmutable, SlotRecord
from ..core import Container, Dependency, DependencyInstance, Provider
from ..core.utils import DependencyDebug


@API.private
class FactoryProvider(Provider):
    def __init__(self):
        super().__init__()
        self.__factories: Dict[Hashable, Factory] = dict()

    def __repr__(self):
        return f"{type(self).__name__}(factories={self.__factories})"

    def clone(self, keep_singletons_cache: bool) -> 'FactoryProvider':
        p = FactoryProvider()
        if keep_singletons_cache:
            factories = {
                k: (f.copy() if f.dependency is not None else f)
                for k, f in self.__factories.items()
            }
        else:
            factories = {
                k: (f.copy(keep_function=False) if f.dependency is not None else f)
                for k, f in self.__factories.items()
            }
        p.__factories = factories
        return p

    def exists(self, dependency: Hashable) -> bool:
        # For now we don't support multiple factories for a single dependency. Neither
        # is sharing the dependency with another provider. Simply because I don't see a
        # use case where it would make sense.
        # Open for discussions though, create an issue if you a use case.
        if isinstance(dependency, Build):
            dependency = dependency.dependency
        return (isinstance(dependency, FactoryDependency)
                and dependency.dependency in self.__factories)

    def maybe_debug(self, build: Hashable) -> Optional[DependencyDebug]:
        dependency_factory = build.dependency if isinstance(build, Build) else build
        if not isinstance(dependency_factory, FactoryDependency):
            return None

        try:
            factory = self.__factories[dependency_factory.dependency]
        except KeyError:
            return None

        return DependencyDebug(
            debug_repr(build),
            singleton=factory.singleton,
            wired=[factory.function] if factory.dependency is None else [],
            dependencies=([factory.dependency]
                          if factory.dependency is not None else []))

    def maybe_provide(self, build: Hashable, container: Container
                      ) -> Optional[DependencyInstance]:
        dependency_factory = build.dependency if isinstance(build, Build) else build
        if not isinstance(dependency_factory, FactoryDependency):
            return None

        try:
            factory = self.__factories[dependency_factory.dependency]
        except KeyError:
            return None

        if factory.function is None:
            f = container.provide(factory.dependency)
            assert f.singleton, "factory dependency is expected to be a singleton"
            factory.function = f.value

        instance = (factory.function(**build.kwargs)
                    if isinstance(build, Build) and build.kwargs
                    else factory.function())

        return DependencyInstance(instance, singleton=factory.singleton)

    def register(self,
                 dependency: Hashable,
                 factory: Union[Callable, Dependency],
                 singleton: bool = True) -> 'FactoryDependency':
        # For now we don't support multiple factories for a single dependency.
        # Simply because I don't see a use case where it would make sense. In
        # Antidote the standard way would be to use `with_kwargs()` to customization
        # Open for discussions though, create an issue if you a use case.
        factory_dependency = FactoryDependency(dependency, weakref.ref(self))
        self._assert_not_duplicate(factory_dependency)

        if isinstance(factory, Dependency):
            self.__factories[dependency] = Factory(dependency=factory.value,
                                                   singleton=singleton)
        elif callable(factory):
            self.__factories[dependency] = Factory(singleton=singleton,
                                                   function=factory)
        else:
            raise TypeError(f"factory must be callable, not {type(factory)!r}.")

        return factory_dependency

    def debug_get_registered_factory(self, dependency: Hashable
                                     ) -> Union[Callable, Dependency]:
        factory = self.__factories[dependency]
        if factory.dependency is not None:
            return Dependency(factory.dependency)
        else:
            return factory.function


@API.private
class FactoryDependency(FinalImmutable):
    __slots__ = ('dependency', '_provider_ref')
    dependency: Hashable
    _provider_ref: 'weakref.ReferenceType[FactoryProvider]'

    def __repr__(self):
        return f"FactoryDependency({self})"

    def __antidote_debug_repr__(self):
        return str(self)

    def __str__(self):
        provider = self._provider_ref()
        dependency = debug_repr(self.dependency)
        if provider is not None:
            factory = provider.debug_get_registered_factory(self.dependency)
            return f"{dependency} @ {debug_repr(factory)}"
        # Should not happen, but we'll try to provide some debug information
        return f"{dependency} @ ???"  # pragma: no cover


@API.private
class Factory(SlotRecord):
    __slots__ = ('singleton', 'function', 'dependency')
    singleton: bool
    function: Callable
    dependency: Any

    def __init__(self,
                 singleton: bool = True,
                 function: Callable = None,
                 dependency: Any = None):
        assert function is not None or dependency is not None
        super().__init__(singleton, function, dependency)

    def copy(self, keep_function: bool = True) -> 'Factory':
        return Factory(self.singleton,
                       self.function if keep_function else None,
                       self.dependency)



from typing import Callable, Dict, Hashable, Optional, Union
from weakref import ref

# @formatter:off
cimport cython
from cpython.object cimport PyObject_Call, PyObject_CallObject
from cpython.ref cimport PyObject

from antidote._providers.service cimport Build
from antidote.core.container cimport (DependencyResult, FastProvider, FLAG_DEFINED,
                                      FLAG_SINGLETON, PyObjectBox, RawContainer)
from .._internal.utils import debug_repr
from ..core import Dependency
from ..core.exceptions import DependencyNotFoundError
from ..core.utils import DependencyDebug
# @formatter:on

cdef extern from "Python.h":
    int PyObject_IsInstance(PyObject *inst, PyObject *cls) except -1
    PyObject*PyDict_GetItem(PyObject *p, PyObject *key)


@cython.final
cdef class FactoryProvider(FastProvider):
    cdef:
        dict __factories
        tuple __empty_tuple
        object __weakref__

    def __init__(self):
        super().__init__()
        self.__factories: Dict[Hashable, Factory] = dict()
        self.__empty_tuple = tuple()

    def __repr__(self):
        return f"{type(self).__name__}(factories={self.__factories})"

    def clone(self, keep_singletons_cache: bool) -> FactoryProvider:
        cdef:
            Factory f
            FactoryProvider p

        p = FactoryProvider()
        if keep_singletons_cache:
            factories = {
                k: (f.copy() if f.dependency is not None else f)
                for k, f in self.__factories.items()
            }
        else:
            factories = {
                k: (f.copy_without_function() if f.dependency is not None else f)
                for k, f in self.__factories.items()
            }
        p.__factories = factories
        return p

    def exists(self, dependency: Hashable) -> bool:
        if isinstance(dependency, Build):
            dependency = dependency.dependency
        return (isinstance(dependency, FactoryDependency)
                and dependency.dependency in self.__factories)

    def maybe_debug(self, build: Hashable) -> Optional[DependencyDebug]:
        cdef:
            Factory factory

        dependency_factory = build.dependency if isinstance(build, Build) else build
        if not isinstance(dependency_factory, FactoryDependency):
            return None

        try:
            factory = self.__factories[dependency_factory.dependency]
        except KeyError:
            return None

        return DependencyDebug(
            debug_repr(build),
            singleton=factory.singleton,
            wired=[factory.function] if factory.dependency is None else [],
            dependencies=([factory.dependency]
                          if factory.dependency is not None else []))

    cdef fast_provide(self,
                      PyObject*dependency,
                      PyObject*container,
                      DependencyResult*result):
        cdef:
            PyObject*factory
            bint is_build_dependency = PyObject_IsInstance(dependency, <PyObject*> Build)
            PyObject*dependency_factory = (<PyObject*> (<Build> dependency).dependency
                                           if is_build_dependency else
                                           dependency)

        if not PyObject_IsInstance(dependency_factory, <PyObject*> FactoryDependency):
            return

        factory = PyDict_GetItem(<PyObject*> self.__factories,
                                 <PyObject*> (
                                     <FactoryDependency> dependency_factory).dependency)
        if factory == NULL:
            return

        if (<Factory> factory).function is None:
            (<RawContainer> container).fast_get(
                <PyObject*> (<Factory> factory).dependency, result)
            if result.flags == 0:
                raise DependencyNotFoundError((<Factory> factory).dependency)
            assert (result.flags & FLAG_SINGLETON) != 0, "factory dependency is expected to be a singleton"
            (<Factory> factory).function = (<PyObjectBox> result.box).obj

        result.flags = FLAG_DEFINED | (<Factory> factory).flags
        if is_build_dependency:
            (<PyObjectBox> result.box).obj = PyObject_Call(
                (<Factory> factory).function,
                self.__empty_tuple,
                (<Build> dependency).kwargs
            )
        else:
            (<PyObjectBox> result.box).obj = PyObject_CallObject(
                (<Factory> factory).function,
                <object> NULL
            )

    def register(self,
                 dependency: Hashable,
                 factory: Union[Callable, Dependency],
                 singleton: bool = True) -> FactoryDependency:
        with self._ensure_not_frozen():
            factory_dependency = FactoryDependency(dependency, ref(self))
            self._raise_if_exists(factory_dependency)

            if isinstance(factory, Dependency):
                self.__factories[dependency] = Factory(singleton,
                                                       dependency=factory.value)
            elif callable(factory):
                self.__factories[dependency] = Factory(singleton, function=factory)
            else:
                raise TypeError(f"factory must be callable, not {type(factory)!r}.")

            return factory_dependency

    def debug_get_registered_factory(self, dependency: Hashable
                                     ) -> Union[Callable, Dependency]:
        cdef:
            Factory factory = self.__factories[dependency]
        if factory.dependency is not None:
            return Dependency(factory.dependency)
        else:
            return factory.function

@cython.final
cdef class FactoryDependency:
    cdef:
        readonly object dependency
        object _provider_ref

    def __init__(self, object dependency, object provider_ref):
        self.dependency = dependency
        self._provider_ref = provider_ref

    def __repr__(self):
        return f"FactoryDependency({self})"

    def __antidote_debug_repr__(self):
        return str(self)

    def __str__(self):
        provider = self._provider_ref()
        dependency = debug_repr(self.dependency)
        if provider is not None:
            factory = provider.debug_get_registered_factory(self.dependency)
            return f"{dependency} @ {debug_repr(factory)}"
        # Should not happen, but we'll try to provide some debug information
        return f"{dependency} @ ???"  # pragma: no cover

@cython.final
cdef class Factory:
    cdef:
        size_t flags
        object function
        object dependency

    def __init__(self,
                 bint singleton,
                 function: Callable = None,
                 dependency: Hashable = None):
        assert function is not None or dependency is not None
        self.flags = FLAG_SINGLETON if singleton else 0
        self.function = function
        self.dependency = dependency

    def __repr__(self):
        return (f"{type(self).__name__}(singleton={self.singleton}, "
                f"function={self.function}, "
                f"dependency={self.dependency})")

    @property
    def singleton(self):
        return self.flags == FLAG_SINGLETON

    def copy(self):
        return Factory(self.singleton,
                       self.function,
                       self.dependency)

    def copy_without_function(self):
        assert self.dependency is not None
        return Factory(self.singleton,
                       None,
                       self.dependency)



import random
import threading
import time

from antidote import factory, Tag, Tagged, world
from antidote.core import Container


class A:
    pass


class B:
    pass


def random_delay(a=0.01, b=None):
    b = b or a
    time.sleep(a + b * random.random())


def delayed_new_class(cls):
    def f() -> cls:
        random_delay()
        return cls()

    return f


def multi_thread_do(target, n_threads=10):
    threads = [threading.Thread(target=target)
               for _ in range(n_threads)]

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join()


def test_container_instantiation_safety():
    with world.test.new():
        n_threads = 10

        build_a = factory(delayed_new_class(A), singleton=True, auto_wire=False)
        build_b = factory(delayed_new_class(B), singleton=False, auto_wire=False)

        singleton_got = []
        non_singleton_got = []

        def worker():
            singleton_got.append(world.get(A @ build_a))
            non_singleton_got.append(world.get(B @ build_b))

        multi_thread_do(worker, n_threads)

        assert len(set(singleton_got)) == 1
        assert len(set(non_singleton_got)) == n_threads


def test_tagged_dependencies_instantiation_safety():
    with world.test.new():
        tag = Tag()
        n_dependencies = 40

        for i in range(n_dependencies):
            factory(delayed_new_class(type(f'Service{i}', (object,), {})),
                    singleton=False,
                    tags=[tag])

        tagged: Tagged = world.get(tag)
        dependencies = []

        def worker():
            for i, dep in enumerate(tagged.values()):
                dependencies.append((i, dep))

        multi_thread_do(worker)

        assert len(set(dependencies)) == n_dependencies
        assert set(enumerate(tagged.values())) == set(dependencies)


# Be sure not have used a fixture to create a new test world as it will
# interfere with this test.
def test_world_safety():
    n_threads = 10
    singletons = []

    def worker():
        with world.test.empty():
            tid = (threading.get_ident(), random.random())
            world.singletons.add("x", tid)
            random_delay()
            singletons.append((tid, tid == world.get("x")))

    multi_thread_do(worker, n_threads)

    assert n_threads == len({tid for (tid, _) in singletons})
    assert all(equal for (_, equal) in singletons)


def test_sate_init_safety():
    from antidote._internal import state
    from antidote._internal.utils import world as world_utils

    old_new_container = world_utils.new_container

    called = 0

    def new_container():
        nonlocal called
        called += 1
        random_delay()
        return old_new_container()

    state.reset()
    world_utils.new_container = new_container

    try:
        multi_thread_do(state.init, n_threads=10)
    finally:
        world_utils.new_container = old_new_container

    assert state.get_container() is not None
    assert called == 1


def test_state_override_safety():
    from antidote._internal import state

    state.init()
    container = state.get_container()

    def create(c):
        assert c is container
        return Container()

    def worker():
        random_delay()
        state.override(create)

    multi_thread_do(worker, n_threads=10)



import pytest

from antidote.core.container import (DependencyInstance, RawContainer,
                                     RawProvider)
from antidote.core.exceptions import DuplicateDependencyError
from antidote.core.utils import DependencyDebug
from antidote.exceptions import (DependencyCycleError, DependencyInstantiationError,
                                 DependencyNotFoundError, FrozenWorldError)
from .utils import DummyFactoryProvider, DummyProvider


class Service:
    def __init__(self, *args):
        pass


class AnotherService:
    def __init__(self, *args):
        pass


class YetAnotherService:
    def __init__(self, *args):
        pass


class ServiceWithNonMetDependency:
    def __init__(self, dependency):
        pass


@pytest.fixture()
def container():
    return RawContainer()


def test_dependency_repr():
    o = object()
    d = DependencyInstance(o, singleton=False)

    assert repr(False) in repr(d)
    assert repr(o) in repr(d)


def test_add_singletons(container: RawContainer):
    x = object()
    y = object()
    container.add_singletons({'x': x, 'y': y})

    assert container.provide('x').value is x
    assert container.provide('x').singleton is True
    assert container.get('y') is y


def test_duplicate_singletons(container: RawContainer):
    x = object()
    container.add_singletons(dict(x=x))

    with pytest.raises(DuplicateDependencyError):
        container.add_singletons(dict(x=object()))

    # did not change singleton
    assert container.get('x') is x


def test_getitem(container: RawContainer):
    container.add_provider(DummyFactoryProvider)
    container.get(DummyFactoryProvider).data = {
        Service: lambda _: Service(),
        ServiceWithNonMetDependency: lambda _: ServiceWithNonMetDependency(),
    }
    container.add_provider(DummyProvider)
    container.get(DummyProvider).data = {'name': 'Antidote'}

    assert isinstance(container.get(Service), Service)
    assert isinstance(container.provide(Service), DependencyInstance)
    assert 'Antidote' == container.get('name')
    assert 'Antidote' == container.provide('name').value

    with pytest.raises(DependencyNotFoundError):
        container.get(object)

    with pytest.raises(DependencyInstantiationError):
        container.get(ServiceWithNonMetDependency)


def test_singleton(container: RawContainer):
    container.add_provider(DummyFactoryProvider)
    container.get(DummyFactoryProvider).data = {
        Service: lambda _: Service(),
        AnotherService: lambda _: AnotherService(),
    }

    service = container.get(Service)
    assert container.get(Service) is service
    assert container.provide(Service).value is service
    assert container.provide(Service).singleton is True

    container.get(DummyFactoryProvider).singleton = False
    another_service = container.get(AnotherService)
    assert container.get(AnotherService) is not another_service
    assert container.provide(AnotherService).value is not another_service
    assert container.provide(AnotherService).singleton is False

    assert container.get(Service) == service


def test_dependency_cycle_error(container: RawContainer):
    container.add_provider(DummyFactoryProvider)
    container.get(DummyFactoryProvider).data = {
        Service: lambda _: Service(container.get(AnotherService)),
        AnotherService: lambda _: AnotherService(container.get(YetAnotherService)),
        YetAnotherService: lambda _: YetAnotherService(container.get(Service)),
    }

    with pytest.raises(DependencyCycleError):
        container.get(Service)

    with pytest.raises(DependencyCycleError):
        container.get(AnotherService)

    with pytest.raises(DependencyCycleError):
        container.get(YetAnotherService)


def test_providers(container: RawContainer):
    x = object()
    container.add_provider(DummyProvider)
    container.get(DummyProvider).data = dict(x=x)
    container.add_provider(DummyFactoryProvider)
    container.get(DummyFactoryProvider).data = dict(y=lambda c: c.get('y'))

    assert len(container.providers) == 2
    for provider in container.providers:
        if isinstance(provider, DummyProvider):
            assert provider.data == dict(x=x)
        else:
            assert 'y' in provider.data


def test_repr_str(container: RawContainer):
    container.add_provider(DummyProvider)
    container.get(DummyProvider).data = {'name': 'Antidote'}
    container.add_singletons({'test': 1})

    assert 'test' in repr(container)
    assert repr(container.get(DummyProvider)) in repr(container)
    assert str(container.get(DummyProvider)) in str(container)


def test_invalid_provider(container: RawContainer):
    with pytest.raises(TypeError):
        container.add_provider(object)

    # Cannot register twice the same kind of provider
    container.add_provider(DummyProvider)
    with pytest.raises(ValueError):
        container.add_provider(DummyProvider)


def test_clone(container: RawContainer):
    container.add_provider(DummyProvider)
    container.get(DummyProvider).data = {'name': 'Antidote'}
    container.add_singletons({'test': object()})

    cloned = container.clone(keep_singletons=True)
    assert cloned.get('test') is container.get('test')
    assert cloned.get(DummyProvider) is not container.get(DummyProvider)

    cloned.add_singletons({'test2': 2})
    with pytest.raises(DependencyNotFoundError):
        container.get("test2")

    cloned = container.clone(keep_singletons=False)
    with pytest.raises(DependencyNotFoundError):
        cloned.get("test")
    assert cloned.get(DummyProvider) is not container.get(DummyProvider)

    cloned.add_singletons({'test2': 2})
    with pytest.raises(DependencyNotFoundError):
        container.get("test2")


def test_freeze(container: RawContainer):
    container.add_provider(DummyProvider)
    container.get(DummyProvider).data = {'name': 'Antidote'}
    container.freeze()

    with pytest.raises(FrozenWorldError):
        container.add_provider(DummyFactoryProvider)

    with pytest.raises(FrozenWorldError):
        container.add_singletons({'test': object()})


def test_ensure_not_frozen(container: RawContainer):
    with container.ensure_not_frozen():
        pass

    container.freeze()

    with pytest.raises(FrozenWorldError):
        with container.ensure_not_frozen():
            pass


def test_provider_property(container: RawContainer):
    container.add_provider(DummyProvider)
    assert container.providers == [container.get(DummyProvider)]


def test_providers_must_properly_clone(container: RawContainer):
    class DummySelf(RawProvider):
        def clone(self, keep_singletons_cache: bool) -> 'RawProvider':
            return self

    container.add_provider(DummySelf)

    with pytest.raises(RuntimeError, match="(?i).*provider.*instance.*"):
        container.clone()


def test_providers_must_properly_clone2(container: RawContainer):
    container.add_provider(DummyProvider)
    p = container.get(DummyProvider)

    class DummyRegistered(RawProvider):
        def clone(self, keep_singletons_cache: bool) -> 'RawProvider':
            return p

    container.add_provider(DummyRegistered)
    with pytest.raises(RuntimeError, match="(?i).*provider.*fresh instance.*"):
        container.clone()


def test_clone_providers(container: RawContainer):
    container.add_provider(DummyProvider)
    data = dict(name='antidote')
    container.get(DummyProvider).data = data

    cloned = container.clone(clone_providers=True)
    assert cloned.get(DummyProvider).data == data

    cloned2 = container.clone(clone_providers=False)
    assert cloned2.get(DummyProvider).data is None


@pytest.mark.filterwarnings("ignore:Debug information")
def test_raise_if_exists(container: RawContainer):
    container.raise_if_exists(1)  # Nothing should happen

    container.add_singletons({1: 10})
    with pytest.raises(DuplicateDependencyError, match=".*singleton.*10.*"):
        container.raise_if_exists(1)

    container.add_provider(DummyProvider)
    container.get(DummyProvider).data = {'name': 'Antidote'}
    with pytest.raises(DuplicateDependencyError, match=".*DummyProvider.*"):
        container.raise_if_exists('name')

    class DummyProviderWithDebug(DummyProvider):
        def debug(self, dependency) -> DependencyDebug:
            return DependencyDebug("debug_info", singleton=True)

    container.add_provider(DummyProviderWithDebug)
    container.get(DummyProviderWithDebug).data = {'hello': 'world'}
    with pytest.raises(DuplicateDependencyError,
                       match=".*DummyProviderWithDebug.*\\ndebug_info"):
        container.raise_if_exists('hello')



# @formatter:off
from typing import Hashable
cimport cython
from cpython.object cimport PyObject, PyObject_CallMethodObjArgs, PyObject_CallObject

from antidote.core.container cimport (Container, DependencyResult, FastProvider,
                                      FLAG_DEFINED, FLAG_SINGLETON, PyObjectBox,
                                      RawContainer, DependencyInstance)
from .._internal.utils import debug_repr
from ..core.exceptions import DependencyNotFoundError
from ..core.utils import DependencyDebug

# @formatter:on

cdef extern from "Python.h":
    int PyObject_IsInstance(PyObject *inst, PyObject *cls) except -1


cdef class Lazy:
    def debug_info(self) -> DependencyDebug:
        raise NotImplementedError()  # pragma: no cover

    cdef fast_lazy_get(self, PyObject*container, DependencyResult*result):
        cdef:
            DependencyInstance dependency_instance
        dependency_instance = self.lazy_get(<Container> container)
        if dependency_instance is not None:
            result.flags = FLAG_DEFINED | (
                FLAG_SINGLETON if dependency_instance.singleton else 0)
            (<PyObjectBox> result.box).obj = dependency_instance.value

    cpdef lazy_get(self, container: Container):
        raise NotImplementedError()  # pragma: no cover

@cython.final
cdef class FastLazyConst(Lazy):
    cdef:
        str name
        object dependency
        str method_name
        object value
        object cast

    def __init__(self, str name, object dependency, str method_name, object value,
                 object cast):
        self.name = name
        self.dependency = dependency
        self.method_name = method_name
        self.value = value
        self.cast = cast

    def debug_info(self) -> DependencyDebug:
        from ..lazy import LazyCall
        if isinstance(self.dependency, LazyCall):
            cls = self.dependency.func
        else:
            cls = self.dependency
        return DependencyDebug(f"Const: {debug_repr(cls)}.{self.name}",
                               singleton=True,
                               dependencies=[self.dependency],
                               wired=[getattr(cls, self.method_name)])

    cdef fast_lazy_get(self, PyObject*container, DependencyResult*result):
        (<RawContainer> container).fast_get(<PyObject*> self.dependency, result)
        if result.flags != 0:
            result.flags = FLAG_DEFINED | FLAG_SINGLETON
            (<PyObjectBox> result.box).obj = PyObject_CallObject(
                self.cast,
                (PyObject_CallMethodObjArgs((<PyObjectBox> result.box).obj,
                                           self.method_name,
                                           <PyObject*> self.value,
                                           NULL),)
            )
        else:
            raise DependencyNotFoundError(self.dependency)

@cython.final
cdef class LazyProvider(FastProvider):
    def clone(self, keep_singletons_cache: bool) -> FastProvider:
        return LazyProvider()

    def exists(self, dependency: Hashable) -> bool:
        return isinstance(dependency, Lazy)

    def maybe_debug(self, dependency: Hashable):
        if isinstance(dependency, Lazy):
            return dependency.debug_info()

    cdef fast_provide(self,
                      PyObject*dependency,
                      PyObject*container,
                      DependencyResult*result):
        if PyObject_IsInstance(dependency, <PyObject*> Lazy):
            return (<Lazy> dependency).fast_lazy_get(container, result)



from typing import Any, Callable, cast, Hashable

from .._compatibility.typing import final
from .._internal import API
from .._internal.utils import debug_repr, FinalImmutable
from ..core import Container, DependencyInstance, StatelessProvider
from ..core.utils import DependencyDebug


@API.private
class Lazy:
    def debug_info(self) -> DependencyDebug:
        raise NotImplementedError()  # pragma: no cover

    def lazy_get(self, container: Container) -> DependencyInstance:
        raise NotImplementedError()  # pragma: no cover


@API.private
@final
class FastLazyConst(FinalImmutable, Lazy):
    __slots__ = ('name', 'dependency', 'method_name', 'value', 'cast')
    name: str
    dependency: object
    method_name: str
    value: object
    cast: Callable[[Any], Any]

    def __init__(self, name: str, dependency, method_name: str, value: object,
                 cast: Callable[[Any], Any]):
        super().__init__(
            name=name,
            dependency=dependency,
            method_name=method_name,
            value=value,
            cast=cast
        )

    def debug_info(self) -> DependencyDebug:
        from ..lazy import LazyCall
        if isinstance(self.dependency, LazyCall):
            cls: type = cast(type, self.dependency.func)
        else:
            cls = cast(type, self.dependency)
        return DependencyDebug(f"Const: {debug_repr(cls)}.{self.name}",
                               singleton=True,
                               dependencies=[self.dependency],
                               # TODO: Would be great if the first argument of the method
                               #       didn't show as unknown as it's always provided.
                               wired=[getattr(cls, self.method_name)])

    def lazy_get(self, container: Container) -> DependencyInstance:
        # TODO: Waiting for a fix: https://github.com/python/mypy/issues/6910
        _cast = cast(Callable[[Any], Any], getattr(self, 'cast'))
        return DependencyInstance(
            _cast(getattr(container.get(self.dependency),
                          self.method_name)(self.value)),
            singleton=True
        )


@API.private
@final
class LazyProvider(StatelessProvider[Lazy]):
    def exists(self, dependency: Hashable) -> bool:
        return isinstance(dependency, Lazy)

    def debug(self, dependency: Lazy) -> DependencyDebug:
        return dependency.debug_info()

    def provide(self, dependency: Lazy, container: Container
                ) -> DependencyInstance:
        return dependency.lazy_get(container)



import threading
from typing import (Any, Dict, Generic, Hashable, Iterable, Iterator, List,
                    Optional, Sequence, Tuple, TypeVar)

from .._compatibility.typing import final
from .._internal import API
from .._internal.utils import debug_repr, short_id
from ..core import Container, DependencyInstance, Provider
from ..core.exceptions import AntidoteError
from ..core.utils import DependencyDebug


@API.public
class Tag:
    """
    Tags are a way to expose a dependency indirectly. Instead of explicitly
    defining a list of dependencies to retrieve, one can just mark those with
    tags and retrieve them.

    The only requirement for a tag is to be an instance of :py:class:`.Tag`.

    .. doctest:: providers_tag_Tag

        >>> from antidote import Tag, Service, world
        >>> tag = Tag()
        >>> class Dummy(Service):
        ...     __antidote__ = Service.Conf(tags=[tag])
        >>> world.get(tag)
        <...Tagged ...>
        >>> # You can retrieve the tags and/or dependencies. Here we take both.
        ... (t, dummy) = list(world.get(tag).items())[0]
        >>> t is tag
        True
        >>> dummy is world.get(Dummy)
        True

    You may create your own subclasses to add additional information on your services.
    You can create Tags will be grouped by the output of :py:meth:`.group`. Retrieved
    dependencies will have their associated tag provided, so you can store information
    in it.

    .. doctest:: providers_tag_Tag_v2

        >>> from antidote import Tag, Service, world, Tagged
        >>> class CustomTag(Tag):
        ...     __slots__ = ('name',)  # __slots__ is recommended
        ...     name: str  # For Mypy
        ...
        ...     def __init__(self, name: str):
        ...         super().__init__(name=name)
        ...
        ...     def group(self):
        ...         return self.name.split("_")[0]
        ...
        >>> class Dummy(Service):
        ...     __antidote__ = Service.Conf(tags=[CustomTag(name="ref_dummy")])
        >>> world.get[Tagged[CustomTag, object]](CustomTag("ref_any"))
        <...Tagged ...>
        >>> (tag, dummy) = list(world.get(CustomTag("ref_any")).items())[0]
        >>> tag.name
        'ref_dummy'
        >>> dummy is world.get(Dummy)
        True

    Note that tags are immutables, well as far it goes in Python ! Tags should be used to
    group dependencies and eventually provides some *constants* information on them. It
    is not intended to store any state in it.

    """
    __slots__ = ()

    def __init__(self, **attrs):
        """
        :py:meth:`.__init__` is the only way to actually set attributes, to be used by
        subclasses.
        """
        for attr, value in attrs.items():
            object.__setattr__(self, attr, value)

    @final
    def __setattr__(self, name, value):
        raise AttributeError(f"{type(self)} is immutable")

    def __antidote_debug_repr__(self):
        group = self.group()
        if group is self:
            group = f"Tag#{short_id(self)}"
        else:
            group = repr(group)
        return f"Tag: {group}"

    def __repr__(self):
        group = self.group()
        if group is self:
            group = f"Tag#{short_id(self)}"
        else:
            group = repr(group)
        return f"{type(self).__name__}(group={group})"

    def group(self):
        """Tags will be grouped by this value. By default it's the tag instance itself."""
        return self


@API.public
class DuplicateTagError(AntidoteError):
    """
    A dependency has multiple times the same tag.
    """

    def __init__(self, dependency: Hashable, existing_tag: Tag):
        self.dependency = dependency
        self.existing_tag = existing_tag

    def __str__(self):
        return f"Dependency {self.dependency} already has a tag {self.existing_tag}"


T = TypeVar('T', bound=Tag)
D = TypeVar('D')


# TODO: Python3.6 does not support inheriting FinalMeta and GenericMeta
#       To be added again once 3.6 support ends.
@API.public
@final
class Tagged(Generic[T, D]):
    """
    Collection containing dependencies and their tags. Dependencies are lazily
    instantiated.
    """

    @API.private  # You're not supposed to create it yourself
    def __init__(self,
                 container: Container,
                 dependencies: Sequence[Hashable],
                 tags: Sequence[T]):
        self.__lock = threading.RLock()
        self.__container = container
        self.__dependencies = list(dependencies)
        self._instances: List[Any] = []
        self._tags = list(tags)

    def __len__(self):
        return len(self._tags)

    def items(self) -> Iterator[Tuple[T, D]]:
        """
        Zips tags and values together.
        """
        return zip(self.tags(), self.values())

    # Mainly here for interface consistency with instances() (instead of _tags)
    def tags(self) -> Iterator[T]:
        """Tags associated with the retrieved dependencies."""
        return iter(self._tags)

    def values(self) -> Iterator[D]:
        """Retrieved dependencies, lazily instantiated."""
        i = -1
        for i, instance in enumerate(self._instances):
            yield instance

        i += 1
        while i < len(self):
            try:
                yield self._instances[i]
            except IndexError:
                with self.__lock:
                    # If not other thread has already added the instance.
                    if i == len(self._instances):
                        self._instances.append(
                            self.__container.get(self.__dependencies[i])
                        )
                yield self._instances[i]
            i += 1

        # Don't need to keep them anymore.
        setattr(self, "__container", None)
        setattr(self, "__dependencies", None)
        setattr(self, "__lock", None)


@API.private
class TagProvider(Provider[Tag]):
    def __init__(self):
        super().__init__()
        self.__tag_to_tagged: Dict[Any, Dict[Hashable, Tag]] = {}

    def __repr__(self):
        return f"{type(self).__name__}(tagged_dependencies={self.__tag_to_tagged})"

    def clone(self, keep_singletons_cache: bool) -> 'TagProvider':
        p = TagProvider()
        p.__tag_to_tagged = self.__tag_to_tagged.copy()
        return p

    def exists(self, dependency: Hashable) -> bool:
        return isinstance(dependency, Tag) and dependency.group() in self.__tag_to_tagged

    def debug(self, dependency: Tag) -> DependencyDebug:
        return DependencyDebug(
            debug_repr(dependency),
            singleton=False,
            dependencies=list(self.__tag_to_tagged[dependency.group()].keys())
        )

    def maybe_provide(self, dependency: Hashable, container: Container
                      ) -> Optional[DependencyInstance]:
        if not isinstance(dependency, Tag):
            return None

        try:
            tagged = self.__tag_to_tagged[dependency.group()]
        except KeyError:
            return None

        return DependencyInstance(
            Tagged(
                container=container,
                dependencies=list(tagged.keys()),
                tags=list(tagged.values())
            ),
            # Whether the returned dependencies are singletons or not is
            # our decision to take.
            singleton=False
        )

    def register(self, dependency: Hashable, *, tags: Iterable[Tag]):
        tags = list(tags)
        for tag in tags:
            if not isinstance(tag, Tag):
                raise TypeError(f"Expecting tag of type Tag, not {type(tag)}")
            if tag.group() not in self.__tag_to_tagged:
                self._assert_not_duplicate(tag)
            # else:
            #   the tag could not be declared elsewhere if other _providers also
            #   check with _assert_not_duplicate and use the freeze lock (which is
            #   enforced @does_not_freeze)

        for tag in tags:
            group = tag.group()
            if group not in self.__tag_to_tagged:
                self.__tag_to_tagged[group] = {dependency: tag}
            elif dependency not in self.__tag_to_tagged[group]:
                self.__tag_to_tagged[group][dependency] = tag
            else:
                raise DuplicateTagError(dependency,
                                        self.__tag_to_tagged[group][dependency])



import pytest

from antidote import factory, implementation, Implementation, Service, world
from antidote._providers import (FactoryProvider, IndirectProvider,
                                 ServiceProvider)
from antidote.exceptions import DependencyInstantiationError


@pytest.fixture(autouse=True)
def test_world():
    with world.test.empty():
        world.provider(ServiceProvider)
        world.provider(FactoryProvider)
        world.provider(IndirectProvider)
        yield


class Interface:
    pass


def test_implementation_class():
    class A(Interface, Implementation):
        pass

    assert world.get(Interface) is world.get(A)


def test_default_implementation():
    class A(Interface, Service):
        pass

    class B(Interface, Service):
        pass

    choice = 'a'

    @implementation(Interface)
    def choose():
        return dict(a=A, b=B)[choice]

    assert world.get(Interface) is world.get(A)
    choice = 'b'
    assert choose() is B
    assert world.get(Interface) is world.get(A)


@pytest.mark.parametrize('singleton,permanent',
                         [(True, True), (True, False), (False, True), (False, False)])
def test_implementation(singleton: bool, permanent: bool):
    choice = 'a'

    class A(Interface, Service):
        __antidote__ = Service.Conf(singleton=singleton)

    class B(Interface, Service):
        __antidote__ = Service.Conf(singleton=singleton)

    @implementation(Interface, permanent=permanent)
    def choose_service():
        return dict(a=A, b=B)[choice]

    assert isinstance(world.get(Interface), A)
    assert (world.get(Interface) is world.get(A)) is singleton

    choice = 'b'
    assert choose_service() == B
    if permanent:
        assert isinstance(world.get(Interface), A)
        assert (world.get(Interface) is world.get(A)) is singleton
    else:
        assert isinstance(world.get(Interface), B)
        assert (world.get(Interface) is world.get(B)) is singleton


def test_implementation_integration():
    x = object()

    with world.test.clone():
        class A(Interface, Service):
            def __init__(self, **kwargs):
                self.kwargs = kwargs

        @implementation(Interface)
        def impl():
            return A.with_kwargs(test=x)

        a = world.get(Interface)
        assert isinstance(a, A)
        assert a.kwargs == dict(test=x)

    with world.test.clone():
        class A(Interface):
            def __init__(self, **kwargs):
                self.kwargs = kwargs

        @factory
        def build_a(**kwargs) -> A:
            return A(**kwargs)

        @implementation(Interface)
        def impl2():
            return A @ build_a.with_kwargs(test=x)

        a = world.get(Interface)
        assert isinstance(a, A)
        assert a.kwargs == dict(test=x)


def test_invalid_implementation_class():
    with pytest.raises(TypeError):
        class A(Implementation):
            pass

    with pytest.raises(TypeError):
        class B(Implementation, Interface):
            pass

    with pytest.raises(TypeError):
        class C(Interface, Implementation, Implementation):
            pass

    with pytest.raises(ValueError):
        class D(Interface, Implementation, abstract=True):
            pass


def test_invalid_implementation():
    with pytest.raises(TypeError):
        implementation(Interface)(1)

    with pytest.raises(TypeError):
        @implementation(Interface)
        class A(Interface):
            pass

    with pytest.raises(TypeError):
        implementation(1)

    class B:
        pass

    world.singletons.add(B, 1)
    world.singletons.add(1, 1)

    with world.test.clone(keep_singletons=True):
        @implementation(Interface)
        def choose():
            return 1

        world.get(1)
        with pytest.raises(DependencyInstantiationError):
            world.get(Interface)

    with world.test.clone(keep_singletons=True):
        @implementation(Interface)
        def choose2():
            return B

        world.get(B)
        with pytest.raises(DependencyInstantiationError):
            world.get(Interface)

    with world.test.clone():
        class C(Service):
            def __init__(self, **kwargs):
                self.kwargs = kwargs

        @implementation(Interface)
        def impl():
            return C.with_kwargs(test=1)

        world.get(C.with_kwargs(test=1))
        with pytest.raises(DependencyInstantiationError):
            world.get(Interface)

    with world.test.clone():
        class D:
            def __init__(self, **kwargs):
                self.kwargs = kwargs

        @factory
        def build_d(**kwargs) -> D:
            return D(**kwargs)

        @implementation(Interface)
        def impl2():
            return D @ build_d.with_kwargs(test=1)

        world.get(D @ build_d.with_kwargs(test=1))
        with pytest.raises(DependencyInstantiationError):
            world.get(Interface)



"""
Utilities used by world, mostly for syntactic sugar.
"""
from typing import Any, Callable, cast, Type, TypeVar, TYPE_CHECKING

from .meta import FinalMeta
from ..._compatibility.typing import final
from ..._internal import API
from ...core.container import RawContainer
from ...core.utils import Dependency

T = TypeVar('T')

if TYPE_CHECKING:
    from mypy_extensions import DefaultArg


@API.private
@final
class WorldGet(metaclass=FinalMeta):
    def __call__(self, dependency: object) -> Any:
        from ..state import get_container
        return get_container().get(dependency)

    def __getitem__(self, tpe: Type[T]
                    ) -> 'Callable[[DefaultArg(object)], T]':
        def f(dependency=None) -> T:
            from ..state import get_container
            if dependency is None:
                dependency = tpe
            return cast(T, get_container().get(dependency))

        return f


@API.private
@final
class WorldLazy(metaclass=FinalMeta):
    def __call__(self, dependency: object) -> Dependency[Any]:
        return Dependency(dependency)

    def __getitem__(self, tpe: Type[T]
                    ) -> 'Callable[[DefaultArg(object)], Dependency[T]]':
        def f(dependency=None) -> Dependency[T]:
            if dependency is None:
                dependency = tpe
            return Dependency(dependency)

        return f


@API.private
def new_container():
    """ default new container in Antidote """

    from ..._providers import (LazyProvider, ServiceProvider, TagProvider,
                               IndirectProvider, FactoryProvider)

    container = RawContainer()
    container.add_provider(FactoryProvider)
    container.add_provider(ServiceProvider)
    container.add_provider(LazyProvider)
    container.add_provider(IndirectProvider)
    container.add_provider(TagProvider)

    return container



import weakref
from typing import Callable, Union

from ._compatibility.typing import final
from ._internal import API
from ._internal.utils import debug_repr, short_id, FinalImmutable
from ._lazy import (LazyCallWithArgsKwargs, LazyMethodCallDependency,
                    LazyMethodCallWithArgsKwargs)
from ._providers import Lazy
from .core import Container, DependencyInstance
from .core.utils import DependencyDebug


@API.public
@final
class LazyCall(FinalImmutable, Lazy):
    """
    Dependency which is the result of the call of the given function with the
    given arguments.

    .. doctest:: helpers_lazy_func

        >>> from antidote import LazyCall, world
        >>> def f(x, y):
        ...     print("Computing {} + {}".format(x, y))
        ...     return x + y
        >>> A = LazyCall(f)(2, y=3)
        >>> world.get(A)
        Computing 2 + 3
        5
        >>> # You may also not provide any arguments
        ... def g():
        ...     print("I'm g")
        >>> B = LazyCall(g)
        >>> world.get(B)
        I'm g

    """
    __slots__ = ('func', 'singleton')
    func: Callable
    singleton: bool

    def __init__(self, func: Callable, *, singleton: bool = True):
        """
        Args:
            func: Function to lazily call, any (keyword-)arguments given by the returned
                :py:class:`~.LazyCall` itself will be propagated.
            singleton: Whether or not this is a singleton or not. If yes, the function
                will only be called once.
        """
        if not callable(func):
            raise TypeError("func is not callable")
        if not isinstance(singleton, bool):
            raise TypeError(f"singleton must be a boolean, not {type(singleton)!r}")
        super().__init__(func=func, singleton=singleton)

    def __call__(self, *args, **kwargs) -> 'LazyCallWithArgsKwargs':
        """
        All argument are passed on to the lazily called function.
        """
        return LazyCallWithArgsKwargs(self.func, self.singleton, args, kwargs)

    def __antidote_debug_repr__(self):
        s = f"Lazy: {debug_repr(self.func)}()"
        if self.singleton:
            s += f"  #{short_id(self)}"
        return s

    def debug_info(self) -> DependencyDebug:
        return DependencyDebug(self.__antidote_debug_repr__(),
                               singleton=self.singleton,
                               wired=[self.func])

    def lazy_get(self, container: Container) -> DependencyInstance:
        return DependencyInstance(self.func(), singleton=self.singleton)


@API.public
@final
class LazyMethodCall(FinalImmutable):
    """
    Similar to :py:class:`~.LazyCall` but adapted to methods within a class definition.
    The class has to be a registered service, as the class instantiation itself is also
    lazy.

    :py:class:`~.LazyMethodCall` has two different behaviors:

    - if retrieved as a class attribute it returns a dependency which can be retrieved
      from Antidote.
    - if retrieved as a instance attribute it returns the result for this
      instance. This makes testing a lot easier as it does not pass through Antidote.

    .. doctest:: helpers_lazy_method

        >>> from antidote import LazyMethodCall, Service, world
        >>> class Api(Service):
        ...     # Name of the method
        ...     conf = LazyMethodCall('query')('/conf')
        ...
        ...     def query(self, url: str = '/status'):
        ...         return f"requested {url}"
        ...
        ...     # or the method itself
        ...     status = LazyMethodCall(query, singleton=False)
        >>> world.get(Api.conf)
        'requested /conf'
        >>> # For ease of use, accessing the dependency through a instance will simply
        ... # call the method without passing through Antidote. Useful for tests typically
        ... Api().status
        'requested /status'

    .. note::

        Check out :py:class:`~.extension.constants.Constants` for simple way
        to declare multiple constants.

    """
    __slots__ = ('method_name', 'singleton', '_cache_attr')
    method_name: str
    singleton: bool
    _cache_attr: str

    def __init__(self, method: Union[Callable, str], *, singleton: bool = True):
        """
        Args:
            method: Method name or the method itself that must be called.
            singleton: Whether or not this is a singleton or not. If yes, the method
                will only be called once.
        """
        if not (callable(method) or isinstance(method, str)):
            raise TypeError("method must be a method or its name")
        if not isinstance(singleton, bool):
            raise TypeError(f"singleton must be a boolean, not {type(singleton)!r}")
        super().__init__(
            method_name=method if isinstance(method, str) else method.__name__,
            singleton=singleton,
            _cache_attr=f"__antidote_dependency_{hex(id(self))}"
        )

    def __call__(self, *args, **kwargs) -> 'LazyMethodCallWithArgsKwargs':
        """
        All argument are passed on to the lazily called method.
        """
        return LazyMethodCallWithArgsKwargs(self.method_name,
                                            self.singleton,
                                            args, kwargs)

    def __str__(self):
        s = f"Lazy Method: {self.method_name}()"
        if self.singleton:
            s += f"  #{short_id(self)}"
        return s

    def __get__(self, instance, owner):
        if instance is None:
            try:
                return getattr(owner, self._cache_attr)
            except AttributeError:
                dependency = LazyMethodCallDependency(self,
                                                      weakref.ref(owner),
                                                      self.singleton)
                setattr(owner, self._cache_attr, dependency)
                return dependency
        return getattr(instance, self.method_name)()
